from __future__ import division, print_function, absolute_import
import numpy as np
from copy import copy
from .saft_forcefield import saft_forcefield
from .constants import kb


class component(object):
    '''
    Creates an object with pure component info

    Parameters
    ----------
    name : str
        Name of the component
    ms : float
        Chain lenght
    sigma : float
        diameter, input in [Ángstrom], stored in [m]
    eps : float
        Mie Potential energy, input in [k], stored in [J]
    lambda_a : float
        Atractive exponent of Mie Potential
    lambda_r : float
        Repulsive exponent of Mie Potential
    eAB : float
        Association Energy, input in [k], stores in [J]
    rcAB : float
        Association rc, input in [sigma], stored in [m]
    rdAB : float
        Association rd, input in [sigma], stored in [m]
    sites : list
        Association sites [Bipolar, Positive, Negative]
    mupol : float
        dipolar moment [Debye]
    npol : float
        numer of polar sites de GV polar term
    ring : float
        geometric factor for ring molecules
        (see Langmuir 2017, 33, 11518-11529, Table I.)
    Tc : float
        Critical temperature [K]
    Pc : float
        Critical Pressure [Pa]
    Zc : float
        critical compresibility factor
    Vc : float
        critical volume [m^3/mol]
    w  : float
        acentric factor
    Mw : float
        molar weight [g/mol]
    cii : list
        polynomial coefficient for influence parameter in SGT [J m^5 / mol^2]

    Methods
    -------
    ci :  evaluates influence parameter polynomial
    saftvrmie_forcefield: use corresponding state principle to correlate
        sigma, eps and lambda_a.
    '''

    def __init__(self, name='None', ms=1., sigma=0., eps=0., lambda_r=12.,
                 lambda_a=6., eAB=0., rcAB=1., rdAB=0.4, sites=[0, 0, 0],
                 mupol=0, npol=0., ring=0., Tc=0., Pc=0., Zc=0., Vc=0.,
                 w=0., Mw=1., cii=0.):

        self.name = name
        self.Tc = Tc  # Critical Temperature in K
        self.Pc = Pc  # Critical Pressure in Pa
        self.Zc = Zc  # Critical compresibility factor
        self.Vc = Vc  # Critical volume in m3/mol
        self.w = w  # Acentric Factor
        self.cii = cii  # Influence factor SGT, list or array
        self.Mw = Mw  # molar weight in g/mol
        self.nc = 1

        # Saft Parameters
        self.ms = ms
        self.sigma = sigma * 1e-10  # meters
        self.eps = eps * kb  # Joule
        self.lambda_a = np.asarray(lambda_a)
        self.lambda_r = np.asarray(lambda_r)
        self.lambda_ar = self.lambda_r + self.lambda_a

        # For ring molecules (see Langmuir 2017, 33, 11518-11529, Table I.)
        self.ring = ring

        # Association Parameters
        self.eAB = eAB * kb  # Joule
        self.rcAB = rcAB * self.sigma  # meters
        self.rdAB = rdAB * self.sigma  # meters
        self.sites = sites

        # Polar parameters
        self.mupol = mupol  # Debye
        self.npol = npol

    def ci(self, T):
        """
        Method that evaluates the polynomial for cii coeffient of SGT
        cii must be in J m^5 / mol^2 and T in K.

        Parameters
        ----------
        T : float
            absolute temperature [K]

        Returns
        -------
        ci : float
            influence parameter at given temperature [J m^5 / mol^2]

        """

        return np.polyval(self.cii, T)

    def saftvrmie_forcefield(self, ms, rhol07):
        """
        saftvrmie_forcefield(ms, rhol07)

        Method that uses corresponding state principle to correlate sigma,
        eps and lambda_r using critical temperature, acentric factor and
        liquid density (Tr=0.7)

        This method requires that critical temperature and acentric factor of
        the fluid has been set.

        This method overwrites the following attributes: ms, sigma, eps,
        lambda_r, lambda_a.

        Parameters
        ----------
        ms : integrer
            number of chain segments
        rhol07: float
            liquid density at reduced temperature = 0.7 [mol/m^3]

        Returns
        -------
        sigma : float,
            size parameter for Mie potential [m]
        eps : float
            well depth for Mie potential [K]
        lambda_r : float
            repulsive exponent for Mie potential [Adim]

        """
        out = saft_forcefield(ms, self.Tc, self.w, rhol07)
        lambda_r, lambda_a, ms, eps, sigma = out
        self.lambda_a = np.asarray(lambda_a)
        self.lambda_r = np.asarray(lambda_r)
        self.lambda_ar = self.lambda_r + self.lambda_a
        self.ms = ms
        self.sigma = sigma
        self.eps = eps
        return sigma, eps/kb, lambda_r


class mixture(object):
    '''
    class mixture
    Creates an object that cointains info about a mixture.

    Parameters
    ----------
    component1 : object
        component created with component class
    component2 : object
        component created with component class

    Attributes
    ----------
    name : list
        Name of the component
    ms : list
        list of chain lenght
    sigma : list
        list of Mie diameter [m]
    eps : list
        List of Mie Potential energy [J]
    lambda_a : list
        List of Atractive exponent of Mie Potential
    lambda_r : list
        List Repulsive exponent of Mie Potential
    eAB : list
        List of Association Energy [J]
    rcAB : list
        List of Association rc [m]
    rdAB : float
        List of Association rd [m]
    sites = list
        Association sites [Bipolar, Positive, Negative]
    mupol : list
        List of dipolar moment [Debye]
    npol : list
        List of numer of polar sites de GV polar term
    Tc : list
        Critical temperature [K]
    Pc : list
        Critical Pressure [Pa]
    Zc : list
        critical compresibility factor
    Vc : list
        critical volume [m^3/mol]
    w  : list
        acentric factor
    cii : list
        polynomial coefficient for influence parameter in SGT [J m^5/mol^2]

    Methods
    -------
    add_component : adds a component to the mixture
    copy: returns a copy of the object
    kij_saft : add kij matrix for SAFT-VR-Mie
    lij_saft : add lij matrix for SAFT-VR-Mie
    ci : computes cij matrix at T for SGT
    '''

    def __init__(self, component1, component2):

        self.names = [component1.name, component2.name]
        self.Tc = [component1.Tc, component2.Tc]
        self.Pc = [component1.Pc, component2.Pc]
        self.Zc = [component1.Zc, component2.Zc]
        self.w = [component1.w, component2.w]
        self.Vc = [component1.Vc, component2.Vc]
        self.cii = [component1.cii, component2.cii]
        self.Mw = [component1.Mw, component2.Mw]
        self.nc = 2

        self.lr = [component1.lambda_r, component2.lambda_r]
        self.la = [component1.lambda_a, component2.lambda_a]
        self.sigma = [component1.sigma, component2.sigma]
        self.eps = [component1.eps, component2.eps]
        self.ms = [component1.ms, component2.ms]
        self.ring = [component1.ring, component2.ring]
        self.eAB = [component1.eAB, component2.eAB]
        self.rc = [component1.rcAB, component2.rcAB]
        self.rd = [component1.rdAB, component2.rdAB]
        self.sitesmix = [component1.sites, component2.sites]

        self.mupol = [component1.mupol, component2.mupol]
        self.npol = [component1.npol, component2.npol]

    def add_component(self, component):
        """
        add_component(component)

        Method that add a component to the mixture

        Parameters
        ----------
        component : object
            pure fluid created with component function
        """
        self.names.append(component.name)
        self.Tc.append(component.Tc)
        self.Pc.append(component.Pc)
        self.Zc.append(component.Zc)
        self.Vc.append(component.Vc)
        self.w.append(component.w)
        self.cii.append(component.cii)
        self.Mw.append(component.Mw)

        self.lr.append(component.lambda_r)
        self.la.append(component.lambda_a)
        self.sigma.append(component.sigma)
        self.eps.append(component.eps)
        self.ms.append(component.ms)
        self.ring.append(component.ring)
        self.eAB.append(component.eAB)
        self.rc.append(component.rcAB)
        self.rd.append(component.rdAB)
        self.sitesmix.append(component.sites)

        self.mupol.append(component.mupol)
        self.npol.append(component.npol)

        self.nc += 1

    def kij_saft(self, kij):
        r"""
        kij_saft(kij)

        Method that adds kij correction for Mie potential well depth.

        .. math::
            \epsilon_{ij} = (1-k_{ij}) \frac{\sqrt{\sigma_i^3 \sigma_j^3}}{\sigma_{ij}^3} \sqrt{\epsilon_i \epsilon_j}

        """
        nc = self.nc
        KIJ = np.asarray(kij)
        shape = KIJ.shape

        isSquare = shape == (nc, nc)
        isSymmetric = np.allclose(KIJ, KIJ.T)

        if isSquare and isSymmetric:
            self.KIJsaft = kij
        else:
            raise Exception('kij matrix is not square or symmetric')

    def lij_saft(self, lij):
        r"""
        lij_saft(lij)

        Method that adds lij correction for association energy.

        .. math::
            \epsilon_{ij}^{AB} = (1 - l_{ij})\sqrt{\epsilon_{ii}^{AB} \epsilon_{jj}^{AB}}

        """

        nc = self.nc
        LIJ = np.asarray(lij)
        shape = LIJ.shape

        isSquare = shape == (nc, nc)
        isSymmetric = np.allclose(LIJ, LIJ.T)

        if isSquare and isSymmetric:
            self.LIJsaft = lij
        else:
            raise Exception('kij matrix is not square or symmetric')

    def ci(self, T):
        """
        Method that computes the matrix of cij interaction parameter for SGT at
        given temperature.

        Parameters
        ----------
        T : float
            absolute temperature [K]

        Returns
        -------
        ci : array_like
            influence parameter matrix at given temperature [J m^5 / mol^2]

        """

        n = len(self.cii)
        ci = np.zeros(n)
        for i in range(n):
            ci[i] = np.polyval(self.cii[i], T)
        self.cij = np.sqrt(np.outer(ci, ci))
        return self.cij

    def copy(self):
        """
        Method that return a copy of the mixture

        Returns
        -------
        mix : object
            returns a copy a of the mixture
        """
        return copy(self)



import numpy as np


# Equation A29 to A33
def kichain(xhix):
    xhix2 = xhix**2
    xhix3 = xhix**3
    xhix4 = xhix2**2
    xhix_1 = (1 - xhix)
    xhix_13 = xhix_1**3
    k0 = -np.log(xhix_1) + (42*xhix - 39*xhix2 + 9*xhix3 - 2*xhix4)/(6*xhix_13)
    k1 = (xhix4 + 6*xhix2 - 12*xhix)/(2*xhix_13)
    k2 = -3*xhix2/(8*xhix_1**2)
    k3 = (-xhix4 + 3*xhix2 + 3*xhix)/(6*xhix_13)
    return np.array([k0, k1, k2, k3])


def dkichain_dxhi00(xhix, dxhix_dxhi00):
    xhix2 = xhix**2
    xhix3 = xhix**3
    xhix4 = xhix2**2
    xhix_1 = (1 - xhix)
    xhix_13 = xhix_1**3

    k0 = -np.log(xhix_1) + (42*xhix - 39*xhix2 + 9*xhix3 - 2*xhix4)/(6*xhix_13)
    k1 = (xhix4 + 6*xhix2 - 12*xhix)/(2*xhix_13)
    k2 = -3*xhix2/(8*xhix_1**2)
    k3 = (-xhix4 + 3*xhix2 + 3*xhix)/(6*xhix_13)

    dk0 = 24 + xhix * (-6 + xhix * (3 - 7 * xhix + xhix**2))
    dk0 *= dxhix_dxhi00 / 3. / (-1 + xhix)**4

    dk1 = 12 + xhix * (2+xhix) * (6 - 6 * xhix + xhix**2)
    dk1 *= -dxhix_dxhi00 / 2. / (-1 + xhix)**4

    dk2 = 3*xhix
    dk2 *= dxhix_dxhi00 / 4. / (-1 + xhix)**3

    dk3 = 3 + xhix * (12 + (-3 + xhix) * (-xhix + xhix**2))
    dk3 *= dxhix_dxhi00 / 6. / (-1 + xhix)**4
    return np.array([[k0, k1, k2, k3],
                     [dk0, dk1, dk2, dk3]])


def d2kichain_dxhi00(xhix, dxhix_dxhi00):
    xhix2 = xhix**2
    xhix3 = xhix**3
    xhix4 = xhix2**2
    xhix_1 = (1 - xhix)
    xhix_13 = xhix_1**3

    k0 = -np.log(xhix_1) + (42*xhix - 39*xhix2 + 9*xhix3 - 2*xhix4)/(6*xhix_13)
    k1 = (xhix4 + 6*xhix2 - 12*xhix)/(2*xhix_13)
    k2 = -3*xhix2/(8*xhix_1**2)
    k3 = (-xhix4 + 3*xhix2 + 3*xhix)/(6*xhix_13)

    dk0 = 24 + xhix * (-6 + xhix * (3 - 7 * xhix + xhix**2))
    dk0 *= dxhix_dxhi00 / 3. / (-1 + xhix)**4

    dk1 = 12 + xhix * (2+xhix) * (6 - 6 * xhix + xhix**2)
    dk1 *= -dxhix_dxhi00 / 2. / (-1 + xhix)**4

    dk2 = 3*xhix
    dk2 *= dxhix_dxhi00 / 4. / (-1 + xhix)**3

    dk3 = 3 + xhix * (12 + (-3 + xhix) * (-xhix + xhix**2))
    dk3 *= dxhix_dxhi00 / 6. / (-1 + xhix)**4

    d2k0 = 3 * (-30 + xhix * (1+xhix) * (4 + xhix))
    d2k0 *= dxhix_dxhi00**2 / 3. / (-1 + xhix)**5

    d2k1 = 12 * (5 - 2 * (-1 + xhix) * xhix)
    d2k1 *= dxhix_dxhi00**2 / 2. / (-1 + xhix)**5

    d2k2 = -3*(1+2*xhix)
    d2k2 *= dxhix_dxhi00**2 / 4. / (-1 + xhix)**4

    d2k3 = 6*(-4 + xhix * (-7 + xhix))
    d2k3 *= dxhix_dxhi00**2 / 6. / (-1 + xhix)**5
    return np.array([[k0, k1, k2, k3],
                     [dk0, dk1, dk2, dk3],
                     [d2k0, d2k1, d2k2, d2k3]])


def dkichain_dx(xhix, dxhix_dx):
    # xhix2 = xhix**2
    # xhix_1 = (1 - xhix)

    dk0 = 24 + xhix * (-6 + xhix * (3 - 7 * xhix + xhix**2))
    dk0 *= dxhix_dx / 3. / (-1 + xhix)**4

    dk1 = 12 + xhix * (2+xhix) * (6 - 6 * xhix + xhix**2)
    dk1 *= -dxhix_dx / 2. / (-1 + xhix)**4

    dk2 = 3*xhix
    dk2 *= dxhix_dx / 4. / (-1 + xhix)**3

    dk3 = 3 + xhix * (12 + (-3 + xhix) * (-xhix + xhix**2))
    dk3 *= dxhix_dx / 6. / (-1 + xhix)**4
    return np.array([dk0, dk1, dk2, dk3]).T


def gdHS(x0i, xhix):
    ks = kichain(xhix)
    xs = np.array([x0i**0, x0i, x0i**2, x0i**3])
    g = np.exp(np.dot(ks, xs))
    return g


def dgdHS_dxhi00(x0i, xhix, dxhix_dxhi00):
    dks = dkichain_dxhi00(xhix, dxhix_dxhi00)
    xs = np.array([x0i**0, x0i, x0i**2, x0i**3])
    dg = np.dot(dks, xs)
    dg[0] = np.exp(dg[0])
    dg[1] *= dg[0]
    return dg


def d2gdHS_dxhi00(x0i, xhix, dxhix_dxhi00):
    d2ks = d2kichain_dxhi00(xhix, dxhix_dxhi00)
    xs = np.array([x0i**0, x0i, x0i**2, x0i**3])
    d2g = np.matmul(d2ks, xs)
    d2g[0] = np.exp(d2g[0])
    d2g[2] += d2g[1]**2
    d2g[2] *= d2g[0]
    d2g[1] *= d2g[0]
    return d2g


def dgdHS_dx(x0i, xhix, dxhix_dx):
    g = gdHS(x0i, xhix)
    dks = dkichain_dx(xhix, dxhix_dx)
    xs = np.array([x0i**0, x0i, x0i**2, x0i**3])
    dg = g * np.matmul(dks, xs)
    return g, dg


def dgdHS_dxxhi(x0i, xhix, dxhix_dxhi00, dxhix_dx):
    g = dgdHS_dxhi00(x0i, xhix, dxhix_dxhi00)
    dks = dkichain_dx(xhix, dxhix_dx)
    xs = np.array([x0i**0, x0i, x0i**2, x0i**3])
    dg = g[0] * np.matmul(dks, xs)
    return g, dg



import numpy as np
from .monomer_aux import xhi_eff, dxhieff_dxhi00, d2xhieff_dxhi00
from .monomer_aux import d3xhieff_dxhi00, dxhieff_dx, dxhieff_dx_dxhi00


# Equation A16
def a1s(xhi00, xhix, xm, cictes, a1vdw):
    # a1s calculation Eq 39
    xhieff = xhi_eff(xhix, cictes)
    ghs = (1 - xhieff/2) / (1 - xhieff)**3
    a1 = a1vdw * ghs * xhi00 * xm
    return a1


def da1s_dxhi00(xhi00, xhix, xm, cictes, a1vdw, dxhix_dxhi00):
    # a1s calculation Eq 39
    xhieff = xhi_eff(xhix, cictes)
    dxhieff = dxhieff_dxhi00(xhix, cictes, dxhix_dxhi00)

    ghs = (1 - xhieff/2) / (1 - xhieff)**3
    a1 = a1vdw * ghs * xhi00 * xm

    da1 = 2 - 3*xhieff + xhieff**2
    da1 += xhi00 * dxhieff * (5 - 2 * xhieff)
    da1 *= a1vdw * xm
    da1 /= 2*(- 1 + xhieff)**4

    return a1, da1


def d2a1s_dxhi00(xhi00, xhix, xm, cictes, a1vdw, dxhix_dxhi00):
    # a1s calculation Eq 39
    xhieff = xhi_eff(xhix, cictes)
    dxhieff = dxhieff_dxhi00(xhix, cictes, dxhix_dxhi00)
    d2xhieff = d2xhieff_dxhi00(xhix, cictes, dxhix_dxhi00)

    ghs = (1 - xhieff/2) / (1 - xhieff)**3
    a1 = a1vdw * ghs * xhi00 * xm

    da1 = 2 - 3*xhieff + xhieff**2
    da1 += xhi00 * dxhieff * (5 - 2 * xhieff)
    da1 *= a1vdw * xm
    da1 /= 2*(- 1 + xhieff)**4

    d2a1 = -2 * (-1 + xhieff) * (-5 + 2*xhieff) * dxhieff
    d2a1 += 6 * xhi00 * (-3 + xhieff) * dxhieff**2
    d2a1 += xhi00 * (-5 + (7 - 2 * xhieff) * xhieff) * d2xhieff
    d2a1 *= a1vdw * xm
    d2a1 /= 2*(- 1 + xhieff)**5

    return a1, da1, d2a1


def d3a1s_dxhi00(xhi00, xhix, xm, cictes, a1vdw, dxhix_dxhi00):

    xhieff = xhi_eff(xhix, cictes)
    dxhieff = dxhieff_dxhi00(xhix, cictes, dxhix_dxhi00)
    d2xhieff = d2xhieff_dxhi00(xhix, cictes, dxhix_dxhi00)
    d3xhieff = d3xhieff_dxhi00(xhix, cictes, dxhix_dxhi00)

    ghs = (1 - xhieff/2) / (1 - xhieff)**3
    a1 = a1vdw * ghs * xhi00 * xm

    da1 = 2 - 3*xhieff + xhieff**2
    da1 += xhi00 * dxhieff * (5 - 2 * xhieff)
    da1 *= a1vdw * xm
    da1 /= 2*(- 1 + xhieff)**4

    d2a1 = -2 * (-1 + xhieff) * (-5 + 2*xhieff) * dxhieff
    d2a1 += 6 * xhi00 * (-3 + xhieff) * dxhieff**2
    d2a1 += xhi00 * (-5 + (7 - 2 * xhieff) * xhieff) * d2xhieff
    d2a1 *= a1vdw * xm
    d2a1 /= 2*(- 1 + xhieff)**5

    d3a1 = 18 * (-3 + xhieff) * (-1 + xhieff) * dxhieff**2
    d3a1 += 12 * xhi00 * (7 - 2 * xhieff) * dxhieff**3
    d3a1 += 18 * xhi00 * (-3 + xhieff) * (-1 + xhieff) * dxhieff * d2xhieff
    d3a1 -= (-1 + xhieff)**2*(-5+2*xhieff)*(3*d2xhieff+xhi00*d3xhieff)

    d3a1 *= a1vdw * xm
    d3a1 /= 2*(- 1 + xhieff)**6

    return a1, da1, d2a1, d3a1


def da1s_dx(xhi00, xhix, xm, ms, cictes, a1vdw, dxhix_dx):
    # a1s calculation Eq 39
    xhieff = xhi_eff(xhix, cictes)
    dxhieff = dxhieff_dx(xhix, cictes, dxhix_dx)

    da1 = np.multiply.outer(ms, (-2 + xhieff) * (-1 + xhieff))
    da1 += (5 - 2 * xhieff) * xm * dxhieff
    da1 *= a1vdw * xhi00
    da1 /= 2*(- 1 + xhieff)**4
    return da1


def da1s_dx_dxhi00(xhi00, xhix, xm, ms, cictes, a1vdw,
                   dxhix_dxhi00, dxhix_dx,  dxhix_dx_dxhi00):

    # a1s calculation Eq 39
    xhieff = xhi_eff(xhix, cictes)
    dxhieffdx = dxhieff_dx(xhix, cictes, dxhix_dx)
    dxhieffdxhi00 = dxhieff_dxhi00(xhix, cictes, dxhix_dxhi00)
    dxhieffdxdxhi00 = dxhieff_dx_dxhi00(xhix, cictes, dxhix_dxhi00,
                                        dxhix_dx, dxhix_dx_dxhi00)

    aux1 = 2 - 3*xhieff + xhieff**2
    aux1 += xhi00 * (5 - 2 * xhieff) * dxhieffdxhi00
    aux1 *= (-1 + xhieff)

    da1 = dxhieffdx
    da1 *= -5 + 7*xhieff - 2*xhieff**2 + 6*xhi00 * (-3 + xhieff)*dxhieffdxhi00
    da1 += xhi00 * dxhieffdxdxhi00 * (-5 + xhieff * (7-2*xhieff))
    da1 *= xm
    da1 += np.multiply.outer(ms, aux1)
    da1 *= a1vdw
    da1 /= 2*(- 1 + xhieff)**5
    return da1
